"""
`Loaders` retrieve data from sources such as files and environment variables
and create a tree to be used by [`load_config`][cfgman.load_config].

A loader is just a function that takes parameter and return a dictionary.
A loader factory is a wrapper over a loader which accepts only one parameter:
a configclass used for the result validation.

`load_config` normally treats all callables as function loader of this kind
and pass them the configclass of the configuration we are loading.

# subpath

In some situations you might want to validate only a subset of a configuration,
in these cases the `subpath` argument of the loader may be handy.

```python
@configclass
class ServerConfig:
    host: str
    port: int

@configclass
class ApplicationConfig:
    name: str = "spam"
    server: ServerConfig

config = load_config(
    ApplicationConfig,
    file_loader(cls=ServerConfig, subpath="server", files=["~/.spam.json"])
)
```

The result loaded by file path will be first validated against the `cls` schema
and then wrapped in a tree at the given subpath.

!!! Note

    We are passing the `cls` parameter, otherwise the validation would be
    performed against the `ApplicationConfig` schema *before* the wrapping. Not what
    we want!

"""

from .env import env_loader as env_loader
from .env import load_env as load_env
from .file import file_loader as file_loader
from .file import load_file as load_file

__all__ = ["env_loader", "load_env", "file_loader", "load_file"]



"""Some internal functions to manipulate trees.

In this context, trees are nested dict where keys are strings and values are
either trees or Any.

"""
from collections.abc import Mapping, Sequence
from typing import Any, NamedTuple

NodePath = Sequence[str]
Node = dict[str, Any]


class ItemPtr(NamedTuple):
    node: Node
    key: str


def ensure_path_prefix(tree: Node, path: NodePath) -> ItemPtr:
    """Ensure the path prefix exists in the tree and return the item pointer.

    NOTE: this function change the tree in-place, adding the required path
    nodes.

    >>> ensure_path_prefix({}, ("a", "b", "c")) == ItemPtr({}, "c")
    True
    >>> tree = {"a": {}}
    >>> ensure_path_prefix(tree, ("a", "b", "c"))
    ItemPtr(node={}, key="c")
    >>> tree == {"a": {"b": {}}}
    True

    """
    if not path:
        raise ValueError("path must contain at least one element.")

    node = tree
    *partial_path, last_key = path
    for key in partial_path:
        node = node.setdefault(key, {})
    return ItemPtr(node, last_key)


def envelop_subpath(tree: Node, subpath: NodePath) -> Node:
    """Return a new tree containing the previous one in a subpath."""

    for k in reversed(subpath):
        tree = {k: tree}
    return tree


def copy(tree: Node) -> Node:
    """Copy a tree.

    This is not a shallow copy of the dict: it recursively copy all contained
    dicts. It also copies all lists in the tree.

    It is not a deep copy either: lists are not recursively copies and we are
    not going to copy any other object as they should be immutable or treated
    as those by the rest of the tree operations.

    """
    tree = tree.copy()
    for k, v in tree.items():
        if isinstance(v, Mapping):
            tree[k] = copy(tree[k])
        elif isinstance(v, list):
            tree[k] = v.copy()
    return tree



"""Environment variable loaders."""

import os
from collections.abc import Callable, Mapping, Sequence
from functools import partial
from typing import Any, TypeVar

from apischema import deserialize

from cfgman.tree import ensure_path_prefix, envelop_subpath

_T = TypeVar("_T")


def load_env(
    cls: type[_T] | None = None,
    *,
    mapping: Mapping[str, str],
    subpath: str | Sequence[str] | None = None,
    validate: bool = False,
    prefix: str = "",
    env: Mapping[str, str] | None = None,
) -> dict[str, Any]:
    """Load a dict from environment variables.

    Given a mapping `envvar` -> `dotted.path`, creates a `dict` where each
    `dotted.path` contains the variable of the `envvar`.

    Missing variables are ignored.

    Examples:
        >>> load_env(
        ...   mapping={"VAR1": "a.b", "VAR2": "a.c", "MISS": "b.c"},
        ...   env={"VAR1": "var1", "VAR2": "var2"}
        ... )
        {'a': {'b': 'var1', 'c': 'var2'}}

    Args:
        cls: class to use for validation.
        mapping: the variable mapping.
        subpath: wrap the result in the subpath of a new `dict`.
        validate: validate the result against the `cls`.
        prefix: prefix to apply to all keys in the mapping.
        env: `dict` to use instead of the environment variables. If `None` use
            [`os.environ`][os.environ].

    Returns:
        A `dict` that can be used as a source for the configuration.
            See [load_config][cfgman.load_config].
    """
    if env is None:
        env = os.environ

    if validate and cls is None:
        raise ValueError("Validation require `cls` being set.")

    # normalize keys to be case-insensitive and add prefix in the mapping keys
    env = {k.upper(): v for k, v in env.items()}
    mapping = {prefix + k.upper(): v for k, v in mapping.items()}

    result: dict[str, Any] = {}

    # iterate all mapping keys and populate the result tree
    for varname, nodepath in mapping.items():
        if varname not in env:
            continue

        node, key = ensure_path_prefix(result, nodepath.split("."))
        node[key] = env[varname]

    if validate:
        deserialize(cls, result, coerce=True)

    # envelop the result in a subpath if required.
    if subpath is None:
        return result

    if isinstance(subpath, str):
        subpath = subpath.split(".")

    return envelop_subpath(result, subpath)


def env_loader(
    *,
    cls: type[_T] | None = None,
    subpath: str | Sequence[str] | None = None,
    mapping: Mapping[str, str],
    validate: bool = False,
    prefix: str = "",
    env: Mapping[str, str] | None = None,
) -> Callable[[type], dict[str, Any]]:
    """Create a new loader for environment variables.

    Accept the same parameters of [`load_env`][cfgman.loaders.env.env_loader]
    but return a callable instead.

    Args:
        cls: class to use for validation.
        mapping: the variable mapping.
        subpath: wrap the result in the subpath of a new `dict`.
        validate: validate the result against the `cls`.
        prefix: prefix to apply to all keys in the mapping.
        env: `dict` to use instead of the environment variables. If `None` use
            [`os.environ`][os.environ].

    Returns:
        A function that can be used as a source for the configuration.
            See [load_config][cfgman.load_config].
    """
    wrapped_func = partial(
        load_env,
        mapping=mapping,
        validate=validate,
        subpath=subpath,
        prefix=prefix,
        env=env,
    )

    if cls is None:
        return wrapped_func
    else:
        return lambda x: wrapped_func(cls)



"""File loaders.

Support multiple file formats and multiple files.
(e.g. default config + user overrides)

"""
import json
from collections.abc import Callable, Iterable, Sequence
from functools import partial
from pathlib import Path
from typing import Any, TypeVar, cast

import apischema

from cfgman.errors import (
    ConfigError,
    NoConfigFileError,
    UnsupportedFileType,
    ValidationError,
)
from cfgman.tree import envelop_subpath
from cfgman.types import FileType

try:
    import yaml

    YAML_EXISTS = True
except KeyError:  # pragma: no cover
    YAML_EXISTS = False

try:
    import tomli

    TOML_EXISTS = True
except KeyError:  # pragma: no cover
    TOML_EXISTS = False


_T = TypeVar("_T")


file_suffixes = {
    ".json": FileType.JSON,
    ".yaml": FileType.YAML,
    ".yml": FileType.YAML,
    ".toml": FileType.TOML,
}


def load_file(
    cls: type[_T],
    *,
    files: Iterable[str | Path] | str | Path,
    subpath: str | Sequence[str] | None = None,
    supported_file_types: Sequence[FileType] | None = None,
    validate: bool = True,
    load_all_files: bool = False,
    load_at_least_one_file: bool = False,
) -> list[dict[str, Any]]:
    """Load a config file or multiple config files.

    Args:
        cls: class used for validation.
        files: either a file path or a list of file paths.
        subpath: wrap the result in the subpath of a new `dict`. Validation is
            done before wrapping.
        supported_file_types: list of supported file types. `None` means all of
            them.
        validate: validate the result.
        load_all_files: if `False` unexisting files are skipped.
        load_at_least_one_file: if `True` an exception is raised when no files
            are loaded.

    Returns:
        A list of dictionary to be used as a source for
        [`load_config`][cfgman.load_config].
    """
    # if supported_file_types is None, then it means all supported files.
    if supported_file_types is None:
        supported_file_types = [FileType.JSON]
        if YAML_EXISTS:
            supported_file_types.append(FileType.YAML)
        if TOML_EXISTS:
            supported_file_types.append(FileType.TOML)

    # check all required libraries are installed.
    if not YAML_EXISTS and FileType.YAML in supported_file_types:
        raise ConfigError("YAML support required but PyYAML is not installed.")
    if not TOML_EXISTS and FileType.TOML in supported_file_types:
        raise ConfigError("TOML support required but tomli is not installed.")

    # load all files
    if isinstance(files, Path | str):
        files = [Path(files)]

    all_paths = [Path(x) for x in files]

    if not load_all_files:
        paths = [x for x in all_paths if x.is_file()]
    else:
        paths = all_paths

    if load_at_least_one_file and not paths:
        raise NoConfigFileError(all_paths)

    contents = [_load_single_file(fname, supported_file_types) for fname in paths]

    # check against the schema
    if validate:
        for fname, content in zip(paths, contents):
            try:
                apischema.deserialize(cls, content, coerce=True)
            except apischema.ValidationError as errors:
                raise ValidationError(str(fname), errors)

    if subpath is None:
        return contents

    if isinstance(subpath, str):
        subpath = subpath.split(".")

    return [envelop_subpath(x, subpath) for x in contents]


def file_loader(
    *,
    cls: type | None = None,
    subpath: str | Sequence[str] | None = None,
    files: Iterable[str | Path] | str | Path,
    supported_file_types: Sequence[FileType] | None = None,
    validate: bool = True,
    load_all_files: bool = False,
    load_at_least_one_file: bool = False,
) -> Callable[[type], list[dict[str, Any]]]:
    """Return a new loader to be used by [`load_config`][cfgman.load_config].

    This is just a wrapper for [`load_file`][cfgman.loaders.file.load_file].

    Args:
        cls: class used for validation.
        files: either a file path or a list of file paths.
        subpath: wrap the result in the subpath of a new `dict`. Validation is
            done before wrapping.
        supported_file_types: list of supported file types. `None` means all of
            them.
        validate: validate the result.
        load_all_files: if `False` unexisting files are skipped.
        load_at_least_one_file: if `True` an exception is raised when no files
            are loaded.

    Returns:
        A function that can be used as a source for the configuration.
            See [load_config][cfgman.load_config].
    """
    wrapped_func = partial(
        load_file,
        subpath=subpath,
        files=files,
        supported_file_types=supported_file_types,
        validate=validate,
        load_all_files=load_all_files,
        load_at_least_one_file=load_at_least_one_file,
    )

    if cls is None:
        return wrapped_func
    else:
        return lambda x: wrapped_func(cls)


def _load_single_file(
    filename: Path, supported_file_types: Sequence[FileType]
) -> dict[str, Any]:
    with open(filename, "rb") as fin:
        try:
            file_type = file_suffixes[filename.suffix]
        except KeyError:
            raise UnsupportedFileType(filename=str(filename))

        if file_type not in supported_file_types:
            raise UnsupportedFileType(file_type=file_type, filename=str(filename))

        match filename.suffix:
            case ".json":
                return cast(dict[str, Any], json.load(fin))
            case ".yaml" | ".yml":
                return cast(dict[str, Any], yaml.safe_load(fin))
            case ".toml":
                return tomli.load(fin)
            case _:
                assert False, "This should never happen."



"""
cfgman is a configuration manager for python.

It provides some utility to:

- combine configuration from different sources (e.g. files, environment
  variables, ...)
- validate the configuration (via apischema)



"""
from collections.abc import Callable, Iterable, Iterator, Mapping, MutableMapping
from dataclasses import MISSING, dataclass, fields, is_dataclass
from typing import Any, TypeVar, cast

from apischema import deserialize
from typing_extensions import dataclass_transform

from .loaders import env_loader, file_loader  # noqa: F401
from .tree import Node, NodePath, copy, ensure_path_prefix

_T = TypeVar("_T")


register = set[type]()
defaults: dict[type, Any] = {}

__all__ = ["MISSING", "configclass", "load_config", "reset", "get_default_config"]


@dataclass_transform()
def configclass(cls: type[_T]) -> type[_T]:
    """Register the class as a config class.

    If the class is not already a [`dataclass`][dataclasses], apply the
    [`dataclass`][dataclasses.dataclass] decorator too.

    Args:
      cls: class to register.

    Returns:
      Either `cls` or `dataclass(cls)`.

    Examples:
        Define a new configclass.

        >>> @configclass
        ... class A:
        ...    x: int
    """
    if not is_dataclass(cls):
        datacls = dataclass(cls)
    else:
        datacls = cls

    register.add(datacls)
    return datacls


def load_config(
    cls: type[_T],
    *sources: dict[str, Any]
    | Callable[[type[_T]], dict[str, Any] | list[dict[str, Any]]],
) -> _T:
    """Load a configuration from multiple sources.

    The order of the sources matters: the first one is the lowest priority one
    (e.g. default values), the last one is the highest priority one (e.g. command
    line parameters).

    Default instances of the `configclasses` in the result tree will be assigned
    if the configuration is loaded correctly.

    Args:
        cls: configuration class to use for both validation and deserialization.
        *sources: configuration sources. A source can be a regular `dict` or a
            [loader][cfgman.loaders].

    Returns:
        An instance of the `configclass` with the data resulting from the merged
            sources.

    """
    if cls not in register:
        raise ValueError("First argument of load_config must be a configclass.")

    unflattened_layers = (x(cls) if callable(x) else x for x in sources)
    layers: list[dict[str, Any]] = []

    for x in unflattened_layers:
        if isinstance(x, Iterable) and not isinstance(x, Mapping):
            layers.extend(x)
        else:
            layers.append(x)

    config = deserialize(cls, _merge_layers(*layers), coerce=True)

    # set default configs
    # NOTE: if the same configclass appears multiple times, the last is the default one.
    for node in _visit_config(config):
        assert (
            is_dataclass(node) and not isinstance(node, type) and type(node) in register
        )
        defaults[type(node)] = node

    return config


def _visit_config(obj: Any) -> Iterator[Any]:
    """Visit the config tree and returns all registered configclasses."""

    if isinstance(obj, type):
        # excludes cases where you have a class/type as a value
        pass
    elif is_dataclass(obj):
        if type(obj) in register:
            yield obj
        for f in fields(obj):
            value = getattr(obj, f.name)
            yield from _visit_config(value)

    elif isinstance(obj, Mapping):
        for value in obj.values():
            yield from _visit_config(value)

    elif isinstance(obj, str | bytes):
        # exclude strings, otherwise they are considered iterables
        pass

    elif isinstance(obj, Iterable):
        for value in obj:
            yield from _visit_config(value)


def _merge_layers(*layers: dict[str, Any]) -> dict[str, Any]:
    """Merge multiple layers into a single one following magic rules.

    Rules:
    - an item is replaced if it is not a dict or a list.
    - dict are merged recusively.
    - lists are joined.
    - nodes containing a MISSING value are pruned.
    """
    head, *tail = layers
    current = copy(head)

    for layer in tail:
        for node_path, node_value in _visit_dict(layer):
            _merge_node_path(current, node_path, node_value)

    _prune_missing_inplace(current)
    return current


def _prune_missing_inplace(d: MutableMapping[str, Any]) -> None:
    """Prune all nodes containing a MISSING values."""

    to_del = []

    for k, v in d.items():
        if v is MISSING:
            to_del.append(k)
        elif isinstance(v, MutableMapping) and v:
            _prune_missing_inplace(v)

    for k in to_del:
        del d[k]


def _visit_dict(root: Node) -> Iterator[tuple[list[str], Any]]:
    """Visit the tree of a dict and yields all node paths.

    Each nested dictionary is visited recursively.
    """
    for key, value in root.items():
        if isinstance(value, dict):
            value = cast(Node, value)
            for node_path, node_value in _visit_dict(value):
                yield [key] + node_path, node_value
        else:
            yield [key], value


def _merge_node_path(tree: Node, path: NodePath, new_value: Any) -> None:
    """Merge the value of a node path with the new value.

    The values is overridden if it is not a list, otherwise the two lists are
    concatenated.
    """
    node, key = ensure_path_prefix(tree, path)

    # don't make any change if the new value is missing
    if new_value is MISSING:
        return

    if isinstance(new_value, list):
        current_value = node.get(key, MISSING)
        if current_value is MISSING:
            current_value = node[key] = []

        if not isinstance(current_value, list):
            raise TypeError(
                f"Setting {'.'.join(path)} has mixed type 'list' and non-'list'."
            )
        else:
            current_value.extend(new_value)
    else:
        node[key] = new_value


def get_default_config(cls: type[_T]) -> _T:
    """Return the default configuration for the given `configclass`.

    Default config instances are created once
    [`load_config`][cfgman.load_config] has been called successfully.

    Note:
        If a class has been instantiated multiple times (because `load_config`
        has been called multiple times or because it has multiple places in the
        configuration) the last instance is considered the default one.

    Args:
        cls: the requested type of the default instance.

    Raises:
        KeyError: if the class is not registered via
            [`configclass`][cfgman.configclass].

    Returns:
        The default instance for `cls`.
    """
    return cast(_T, defaults[cls])


def reset() -> None:
    """Reset the default configclasses.

    *Warning*: this is an utility function for tests, you hardly need to use
    this in your app.
    """
    defaults.clear()



from pathlib import Path

import apischema

from cfgman.types import FileType


class ConfigError(Exception):
    """cfgman configuration error."""


class NoConfigFileError(Exception):
    def __init__(self, paths: list[Path]):
        self._paths = paths

    def __str__(self) -> str:
        return (
            f"No config files in the following paths:"
            f" {', '.join(map(str, self._paths))}"
        )


class ValidationError(Exception):
    def __init__(self, filename: str, errors: apischema.ValidationError):
        self._filename = filename
        self._errors = errors

    def __str__(self) -> str:
        filename = self._filename
        errors = self._errors
        return f"Errors found in {filename}: {errors}"


class UnsupportedFileType(Exception):
    """The imported file is not supported."""

    def __init__(
        self, *, file_type: FileType | None = None, filename: str | None = None
    ):
        if file_type is None and filename is not None:
            msg = f"Cannot detect type for file '{filename}'"
        elif filename is None and file_type is not None:
            msg = f"{file_type.name} is not supported."
        elif file_type is not None:
            msg = f"{file_type.name} is not supported for file '{filename}'."
        else:
            msg = "Unsupported file type."
        super().__init__(msg)

        self.file_format = file_type
        self.filename = filename



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cfgman', 'cfgman.loaders']

package_data = \
{'': ['*']}

install_requires = \
['apischema>=0.17.5,<0.18.0', 'pytest>=7.1.3,<8.0.0']

extras_require = \
{'all': ['PyYAML>=6.0,<7.0', 'tomli>=2.0.1,<3.0.0'],
 'toml': ['tomli>=2.0.1,<3.0.0'],
 'yaml': ['PyYAML>=6.0,<7.0']}

setup_kwargs = {
    'name': 'cfgman',
    'version': '0.1.0',
    'description': 'Configuration manager made easy.',
    'long_description': '# cfgman\n\nA configuration manager for Python application.\n\n## Features\n\n- Configure from files (yaml, toml, json).\n- Environment variables.\n- Custom variables (e.g. command line arguments).\n- Type annotations.\n\n## Usage\n\n```python\nfrom cfgman import (\n    configclass,\n    load_config,\n    get_default_config,\n    env_loader,\n    file_loader,\n    MISSING,\n)\nfrom apischema.metadata import conversion\n\n\ndef from_isoformat(s: str) -> datetime:\n    return datetime.fromisoformat(s)\n\n\ndef to_isoformat(d: datetime) -> str:\n    return d.isoformat()\n\n\n@configclass\nclass WebServerConfig:\n    host: str = "localhost"\n    port: int = 80\n    timeout: int = 30\n\n\n@configclass\nclass Config:\n    web: WebServerConfig  # nest configs\n    start: datetime = field(\n        default_factory=lambda: datetime.now(datetime.utc),\n        metadata=conversion(from_isoformat, to_isoformat),\n    )\n\n\nargparser = ArgumentParser(Config)\nargparser.add_argument(\n    "-c",\n    "--config",\n    dest="config_files",\n    metavar="FILENAME",\n    action="append",\n    default=["~/.config/myapp/config.yaml"],\n)\nargparser.add_argument("--host", metavar="HOSTNAME", default=MISSING)\nargparser.add_argument("-p", "--port", metavar="PORT", type=int, default=MISSING)\nargparser.add_argument("-s", "--start", metavar="ISODATETIME", default=MISSING)\nargs = argparser.parse_args()\n\nconfig = load_config(\n    Config,\n    file_loader(\n        files=args.config_files, supported_formats=(FileFormat.YAML, FileFormat.TOML)\n    ),\n    env_loader(\n        prefix="APP_",\n        mapping={"HOST": "web.host", "PORT": "web.port"},\n    ),\n    {\n        "web": {\n            "host": args.host,\n            "port": args.port,\n        },\n        "start": args.start,\n    },\n)\n\ntime.sleep((datetime.now(datetime.utc) - config.start).total_seconds)\n\n\nasync def run_server():\n    # you can retrieve your module specific config without depending on the rest of the configuration\n    webconfig = get_default_config(WebServerConfig)\n    server = start_server(webconfig.host, webconfig.port)\n    await server.serve_forever()\n\n\nasync def request_handler(request):\n    experiment = select_experiment()\n    with replace_default(WebServerConfig, changes=experiment.changes):\n        webconfig = get_default_config(WebServerConfig)\n        response = await prepare_response(\n            headers={"X-Experiment-Name": webconfig.experiment}\n        )\n```\n',
    'author': 'Maurizio Sambati',
    'author_email': 'maurizio@skicelab.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://duilio.github.io/cfgman/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
